/*
 * EEPROM calibration block retreival code for fmc-tdc.
 *
 * Copyright (C) 2013 CERN (www.cern.ch)
 * Author: Tomasz Włostowski <tomasz.wlostowski@cern.ch>
 * Author: Alessandro Rubini <rubini@gnudd.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2 as published by the Free Software Foundation or, at your
 * option, any later version.
 */

#include <linux/platform_device.h>
#include <linux/time.h>
#include <linux/stat.h>

#include "fmc-tdc.h"

/* dummy calibration data - used in case of empty/corrupted EEPROM */
static struct ft_calibration default_calibration = {
	{0, 86, 609, 572, 335},	/* zero_offset */
	43343			/* vcxo_default_tune */
};


/* This is the only thing called by outside */
int ft_handle_calibration(struct fmctdc_dev *ft, struct ft_calibration *calib)
{
	struct device *d = &ft->pdev->dev;
	int i;
	u32 raw_calib[7];

	if (calib) {
		calib->zero_offset[0] = 0;
		for (i = FT_CH_1 + 1; i < FT_NUM_CHANNELS; i++)
			calib->zero_offset[i] =
			  le32_to_cpu(raw_calib[i - 1]) / 100
			  - calib->zero_offset[0];

		calib->vcxo_default_tune = le32_to_cpu(raw_calib[4]);
		calib->calibration_temp = le32_to_cpu(raw_calib[5]);
		calib->wr_offset = le32_to_cpu(raw_calib[6]) / 100;
		ft->calib = *calib;
	} else {
		dev_err(d,
			"Failed to read calibration EEPROM. Using default.\n");
		ft->calib = default_calibration;
	}

	if (ft->verbose) {
		/* Print verbose messages */
		for (i = 0; i < ARRAY_SIZE(ft->calib.zero_offset); i++)
			dev_info(d, "calib: zero_offset[%i] = %i ps\n",
				 i, ft->calib.zero_offset[i]);

		dev_info(d, "calib: vcxo_default_tune %i\n",
			 ft->calib.vcxo_default_tune);
		dev_info(d, "calib: wr offset = %i ps\n",
			 ft->calib.wr_offset);
	}


	return 0;
}


static ssize_t ft_write_eeprom(struct file *file, struct kobject *kobj,
			       struct bin_attribute *attr,
			       char *buf, loff_t off, size_t count)
{
	struct device *dev = container_of(kobj, struct device, kobj);
	struct platform_device *pdev = to_platform_device(dev);
	struct fmctdc_dev *ft = platform_get_drvdata(pdev);
	struct ft_calibration *calib = (struct ft_calibration *)(buf + 0x240);
	int ret;

	if (off >= (4 * 1024)) {
		/*
		 * We do care only about the data in the first 4K of
		 * the eeprom. Just, acknowledge any other page
		 */
		return count;
	}

	if (count < 0x240 + sizeof(struct ft_calibration)) {
		dev_err(dev, "Invalid eeprom size\n");
		return -EINVAL;
	}
	ret = ft_handle_calibration(ft, calib);
	return ret ? ret : count;
}
struct bin_attribute dev_attr_eeprom = {
	.attr = {
		.name = "eeprom",
		.mode = S_IWUSR,
	},
	.size = (8 * 1024), /* 8 KiB */
	.write = ft_write_eeprom,
};
