/*
 * Main fmc-tdc driver module.
 *
 * Copyright (C) 2012-2013 CERN (www.cern.ch)
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 * Author: Alessandro Rubini <rubini@gnudd.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2 as published by the Free Software Foundation or, at your
 * option, any later version.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/bitops.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/io.h>
#include <linux/platform_device.h>

#include <linux/zio.h>
#include <linux/zio-trigger.h>

#include "fmc-tdc.h"
#include "hw/tdc_regs.h"

/* Module parameters */
static int ft_verbose;
module_param_named(verbose, ft_verbose, int, 0444);
MODULE_PARM_DESC(verbose, "Print a lot of debugging messages.");

struct ft_memory_ops memops = {
	.read = NULL,
	.write = NULL,
};


static int ft_reset_core(struct fmctdc_dev *ft)
{
	struct resource *r;
	uint32_t val, shift = 0;
	void *addr;

	addr = ft->ft_carrier_base + TDC_REG_CARRIER_RST;
	switch (ft->pdev->id_entry->driver_data) {
	case TDC_VER_SPEC:
		shift = -1;
		break;
	default:
		break;
	}

	dev_dbg(&ft->pdev->dev, "Un-resetting FMCs...\n");

	/* FMC slot counting starts from 1 */
	r = platform_get_resource(ft->pdev, IORESOURCE_BUS, TDC_BUS_FMC_SLOT);

	/* Reset - reset bits are shifted by 1 */
	/* FIXME pdev->id is not correct because we mayhave different TDCs
	   in slot 0 on different carriers */
	ft_iowrite(ft, ~(1 << (r->start + shift)), addr);

	udelay(5000);

	val = ft_ioread(ft, addr);
	val |= (1 << (r->start + shift));

	/* Un-Reset */
	ft_iowrite(ft, val, addr);

	return 0;
}

static int ft_init_channel(struct fmctdc_dev *ft, int channel)
{
	struct ft_channel_state *st = &ft->channels[channel - 1];

	st->expected_edge = 1;
	return 0;
}


int ft_enable_termination(struct fmctdc_dev *ft, int channel, int enable)
{
	struct ft_channel_state *st;
	uint32_t ien;

	if (channel < FT_CH_1 || channel > FT_NUM_CHANNELS)
		return -EINVAL;

	st = &ft->channels[channel - 1];

	ien = ft_readl(ft, TDC_REG_INPUT_ENABLE);

	if (enable)
		ien |= (1 << (channel - 1));
	else
		ien &= ~(1 << (channel - 1));

	ft_writel(ft, ien, TDC_REG_INPUT_ENABLE);

	if (enable)
		set_bit(FT_FLAG_CH_TERMINATED, &st->flags);
	else
		clear_bit(FT_FLAG_CH_TERMINATED, &st->flags);

	return 0;
}

void ft_enable_acquisition(struct fmctdc_dev *ft, int enable)
{
	uint32_t ien;

	ien = ft_readl(ft, TDC_REG_INPUT_ENABLE);
	if (enable) {
		/* Enable TDC acquisition */
		ft_writel(ft, ien | TDC_INPUT_ENABLE_CH_ALL | TDC_INPUT_ENABLE_FLAG,
			  TDC_REG_INPUT_ENABLE);
		/* Enable ACAM acquisition */
		ft_writel(ft, TDC_CTRL_EN_ACQ, TDC_REG_CTRL);
	} else {
		/* Disable ACAM acquisition */
		ft_writel(ft, TDC_CTRL_DIS_ACQ, TDC_REG_CTRL);
		/* Disable TDC acquisition */
		ft_writel(ft, ien & ~(TDC_INPUT_ENABLE_CH_ALL | TDC_INPUT_ENABLE_FLAG),
			  TDC_REG_INPUT_ENABLE);
	}
}


static int ft_channels_init(struct fmctdc_dev *ft)
{
	int i, ret;

	for (i = FT_CH_1; i <= FT_NUM_CHANNELS; i++) {
		ret = ft_init_channel(ft, i);
		if (ret < 0)
			return ret;
		/* termination is off by default */
		ft_enable_termination(ft, i, 0);
	}
	return 0;
}

static void ft_channels_exit(struct fmctdc_dev *ft)
{
	return;
}

struct ft_modlist {
	char *name;

	int (*init)(struct fmctdc_dev *);
	void (*exit)(struct fmctdc_dev *);
};

static struct ft_modlist init_subsystems[] = {
	{"acam-tdc", ft_acam_init, ft_acam_exit},
	{"onewire", ft_onewire_init, ft_onewire_exit},
	{"time", ft_time_init, ft_time_exit},
	{"channels", ft_channels_init, ft_channels_exit},
	{"zio", ft_zio_init, ft_zio_exit}
};

static int ft_resource_validation(struct platform_device *pdev)
{
	struct resource *r;

	r = platform_get_resource(pdev, IORESOURCE_IRQ, TDC_IRQ);
	if (!r) {
		dev_err(&pdev->dev,
			"The TDC needs an interrupt number\n");
		return -ENXIO;
	}

	r = platform_get_resource(pdev, IORESOURCE_MEM, TDC_MEM_BASE);
	if (!r) {
		dev_err(&pdev->dev,
			"The TDC needs base address\n");
		return -ENXIO;
	}

	r = platform_get_resource(pdev, IORESOURCE_MEM, TDC_CARR_MEM_BASE);
	if (!r) {
		dev_err(&pdev->dev,
			"The TDC needs the carrier base address\n");
		return -ENXIO;
	}

	r = platform_get_resource(pdev, IORESOURCE_BUS, TDC_BUS_FMC_SLOT);
	if (!r) {
		dev_err(&pdev->dev,
			"The TDC needs to be assigned to an FMC slot\n");
		return -ENXIO;
	}

	return 0;
}

/**
 * probe and remove are called by the FMC bus core
 */
int ft_probe(struct platform_device *pdev)
{
	struct ft_modlist *m;
	struct fmctdc_dev *ft;
	struct device *dev = &pdev->dev;
	struct resource *r;
	int i, ret, err;

	err = ft_resource_validation(pdev);
	if (err)
		return err;

	/* Assign IO operation */
	switch (pdev->id_entry->driver_data) {
	case TDC_VER_SPEC:
		memops.read = ioread32;
		memops.write = iowrite32;
		break;
	case TDC_VER_SVEC:
		memops.read = ioread32be;
		memops.write = iowrite32be;
		break;
	default:
		dev_err(&pdev->dev, "Unknow version %lu\n",
			pdev->id_entry->driver_data);
		return -EINVAL;
	}

	ft = kzalloc(sizeof(struct fmctdc_dev), GFP_KERNEL);
	if (!ft) {
		dev_err(dev, "can't allocate device\n");
		return -ENOMEM;
	}
	platform_set_drvdata(pdev, ft);
	ft->pdev = pdev;

	ft->verbose = ft_verbose;

	/* Now use SDB to find the base addresses */
	r = platform_get_resource(pdev, IORESOURCE_MEM, TDC_CARR_MEM_BASE);
	ft->ft_carrier_base = ioremap(r->start, resource_size(r));
	r = platform_get_resource(pdev, IORESOURCE_MEM, TDC_MEM_BASE);
	ft->ft_base = ioremap(r->start, resource_size(r));
	ft->ft_core_base = ft->ft_base + TDC_MEZZ_CORE_OFFSET;
	ft->ft_irq_base = ft->ft_base + TDC_MEZZ_EIC_OFFSET;
	ft->ft_owregs_base = ft->ft_base + TDC_MEZZ_ONEWIRE_OFFSET;
	ft->ft_buffer_base = ft->ft_base + TDC_MEZZ_MEM_OFFSET;

	if (ft_verbose) {
		dev_info(dev,
			 "Base addrs: carr %p, core %p, irq %p, 1wire %p, buffer/DMA %p\n",
			 ft->ft_carrier_base,
			 ft->ft_core_base, ft->ft_irq_base,
			 ft->ft_owregs_base, ft->ft_buffer_base);
	}

		ret = ft_reset_core(ft);
	if (ret < 0)
		return ret;

	spin_lock_init(&ft->lock);

	/* Retrieve calibration from the eeprom, and validate */
	ret = ft_handle_calibration(ft, NULL);
	if (ret < 0)
		return ret;

	/* init all subsystems */
	for (i = 0, m = init_subsystems; i < ARRAY_SIZE(init_subsystems);
	     i++, m++) {
		ret = m->init(ft);
		if (ret < 0)
			goto err;
	}

	ret = ft_irq_init(ft);
	if (ret < 0)
		goto err;

	ret = device_create_bin_file(&ft->pdev->dev, &dev_attr_eeprom);
	if (ret)
		goto err_bin;

	ft_enable_acquisition(ft, 1);
	ft->initialized = 1;
	ft->sequence = 0;

	return 0;

err_bin:
	ft_irq_exit(ft);
err:
	while (--m, --i >= 0)
		if (m->exit)
			m->exit(ft);
	return ret;
}

int ft_remove(struct platform_device *pdev)
{
	struct ft_modlist *m;
	struct fmctdc_dev *ft = platform_get_drvdata(pdev);

	int i = ARRAY_SIZE(init_subsystems);

	if (!ft->initialized)
		return 0;	/* No init, no exit */

	device_remove_bin_file(&ft->pdev->dev, &dev_attr_eeprom);
	ft_enable_acquisition(ft, 0);
	ft_irq_exit(ft);

	while (--i >= 0) {
		m = init_subsystems + i;
		if (m->exit)
			m->exit(ft);
	}

	return 0;
}


static const struct platform_device_id ft_id[] = {
	{
		.name = "fmc-tdc-spec",
		.driver_data = TDC_VER_SPEC,
	}, {
		.name = "fmc-tdc-svec",
		.driver_data = TDC_VER_SVEC,
	},

	/* TODO we should support different version */
};

static struct platform_driver ft_platform_driver = {
	.driver = {
		.name = KBUILD_MODNAME,
	},
	.probe = ft_probe,
	.remove = ft_remove,
	.id_table = ft_id,
};

static int ft_init(void)
{
	int ret;

	ret = ft_zio_register();
	if (ret < 0)
		return ret;

	ret = platform_driver_register(&ft_platform_driver);
	if (ret < 0) {
		ft_zio_unregister();
		return ret;
	}
	return 0;
}

static void ft_exit(void)
{
	platform_driver_unregister(&ft_platform_driver);
	ft_zio_unregister();
}

module_init(ft_init);
module_exit(ft_exit);

MODULE_VERSION(GIT_VERSION);
MODULE_LICENSE("GPL and additional rights");	/* LGPL */

ADDITIONAL_VERSIONS;
